#ifndef DISKCACHE__QUERY_HH
#define DISKCACHE__QUERY_HH

#include "config.h"

#include <iostream>

#include <list>
#include <map>
#include <stdexcept>
#include <string>

#include "general/types.hh"

namespace diskCache
{
  class FilenameHandler;

  class Query
  {
  public:
    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    typedef INT_2U bit_flag_type;
    //-------------------------------------------------------------------
    /// \brief Type used to represent GPS seconds.
    //-------------------------------------------------------------------
    typedef INT_4U gps_time_type;

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    typedef FilenameHandler* filename_handler_type;

    typedef std::map< gps_time_type, gps_time_type >	time_range_type;

    enum {
      CRITERIA_BIT_NONE       		= 0x0000,
      CRITERIA_BIT_IFO        		= 0x0001,
      CRITERIA_BIT_TYPE       		= 0x0002,
      CRITERIA_BIT_START_TIME 		= 0x0004,
      CRITERIA_BIT_END_TIME   		= 0x0008,
      CRITERIA_BIT_EXT        		= 0x0010,
      CRITERIA_BIT_FILENAME_HANDLER     = 0x0020,
      CRITERIA_BIT_IFO_TYPE		= 0x0040,
      CRITERIA_BIT_IFO_TYPE_LIST	= 0x0080,
      CRITERIA_BIT_IFO_TYPE_ABSOLUTE	= 0x0100,
      CRITERIA_BIT_MOUNT_POINT		= 0x0200,
      CRITERIA_BIT_TIME_RANGE		= 0x0400
    };

    typedef std::list< std::string >	ifo_type_list_type;


    //-------------------------------------------------------------------
    /// \brief Copy constructor
    //-------------------------------------------------------------------
    Query( const Query& Source );

    //-------------------------------------------------------------------
    /// \brief Default constructor
    //-------------------------------------------------------------------
    Query( );

    //-------------------------------------------------------------------
    /// \brief Destructor
    //-------------------------------------------------------------------
    ~Query( );

    bit_flag_type Criteria( ) const;

    gps_time_type EndTime( ) const;

    void EndTime( gps_time_type Value );

    const std::string&  Extension( ) const;

    void Extension( const std::string& Value );

    filename_handler_type FNHandler( ) const;

    void FNHandler( filename_handler_type Value );

    const std::string& IFO( ) const;

    void IFO( const std::string& Value );

    const std::string& IFOType( ) const;

    void IFOType( const std::string& Value );

    const std::string& IFOTypeAbsolute( ) const;

    void IFOTypeAbsolute( const std::string& Value );

    const ifo_type_list_type& IFOTypeList( ) const;

    void IFOTypeList( const std::string& Value );

    const std::string& MountPoint( ) const;

    void MountPoint( const std::string& Value );

    time_range_type& TimeRange( );

    const time_range_type& TimeRange( ) const;

    void TimeRange( const gps_time_type Start,
		    const gps_time_type End );

    void TimeRange( time_range_type& NewRange );

    const std::string& Type( ) const;

    void Type( const std::string& Value );

    gps_time_type StartTime( ) const;

    void StartTime( gps_time_type Value );

  private:

    bit_flag_type	m_search_criteria;

    //-------------------------------------------------------------------
    /// \brief IFO on which to match
    //-------------------------------------------------------------------
    std::string		m_ifo;
    //-------------------------------------------------------------------
    /// \brief IFO on which to match
    //-------------------------------------------------------------------
    std::string		m_ifo_type;
    //-------------------------------------------------------------------
    /// \brief IFO on which to match
    //-------------------------------------------------------------------
    ifo_type_list_type	m_ifo_type_list;
    //-------------------------------------------------------------------
    /// \brief Mount point on which to match
    //-------------------------------------------------------------------
    std::string		m_mount_point;
    //-------------------------------------------------------------------
    /// \brief Description on which to match
    //-------------------------------------------------------------------
    std::string		m_type;
    //-------------------------------------------------------------------
    /// \brief Start time on which to match
    //-------------------------------------------------------------------
    gps_time_type	m_time_start;
    //-------------------------------------------------------------------
    /// \brief End time on which to match
    //-------------------------------------------------------------------
    gps_time_type	m_time_end;
    //-------------------------------------------------------------------
    /// \brief File extention on which to match
    //-------------------------------------------------------------------
    std::string		m_ext;
    //-------------------------------------------------------------------
    /// \brief Filename handler on which to match
    //-------------------------------------------------------------------
    filename_handler_type	m_filename_handler;

    time_range_type*		m_time_range;
    Query*			m_time_range_parent;

    void normalize( );

    void set_criteria_bit( bit_flag_type Bit );
  };

  inline void Query::
  normalize( )
  {
    //-------------------------------------------------------------------
    // Check for the ifo type restrictions
    //-------------------------------------------------------------------
    register bit_flag_type	criteria = m_search_criteria;

    if ( criteria & ( CRITERIA_BIT_IFO
		      | CRITERIA_BIT_TYPE
		      | CRITERIA_BIT_IFO_TYPE
		      | CRITERIA_BIT_IFO_TYPE_ABSOLUTE
		      | CRITERIA_BIT_IFO_TYPE_LIST ) )
    {
      if ( criteria & ( CRITERIA_BIT_IFO
			| CRITERIA_BIT_TYPE ) )
      {
	m_ifo_type.clear( );
	m_ifo_type_list.clear( );
      }
      else if ( criteria & ( CRITERIA_BIT_IFO_TYPE
			     | CRITERIA_BIT_IFO_TYPE_ABSOLUTE ) )
      {
	m_ifo.clear( );
	m_type.clear( );
	m_ifo_type_list.clear( );
      }
      else if ( criteria & CRITERIA_BIT_IFO_TYPE_LIST )
      {
	m_ifo.clear( );
	m_type.clear( );
	m_ifo_type.clear( );
      }
    }
  }

  inline void Query::
  set_criteria_bit( bit_flag_type Bit )
  {
    switch( Bit )
    {
    case CRITERIA_BIT_IFO:
    case CRITERIA_BIT_TYPE:
      m_search_criteria =
	( ( m_search_criteria & 
	    //-----------------------------------------------------------
	    // Clear out the old value
	    //-----------------------------------------------------------
	    ~( CRITERIA_BIT_IFO_TYPE
	       | CRITERIA_BIT_IFO_TYPE_ABSOLUTE
	       | CRITERIA_BIT_IFO_TYPE_LIST
	       ) )
	  //-------------------------------------------------------------
	  // Set the new status
	  //-------------------------------------------------------------
	  | Bit );
      m_search_criteria |= Bit;
      break;
    case CRITERIA_BIT_IFO_TYPE:
    case CRITERIA_BIT_IFO_TYPE_ABSOLUTE:
    case CRITERIA_BIT_IFO_TYPE_LIST:
      m_search_criteria =
	( ( m_search_criteria &
	    //-----------------------------------------------------------
	    // Clear out the old value
	    //-----------------------------------------------------------
	    ~( CRITERIA_BIT_IFO
	       | CRITERIA_BIT_TYPE
	       | CRITERIA_BIT_IFO_TYPE
	       | CRITERIA_BIT_IFO_TYPE_ABSOLUTE
	       | CRITERIA_BIT_IFO_TYPE_LIST
	       ) )
	  //-------------------------------------------------------------
	  // Set the new status
	  //-------------------------------------------------------------
	  | Bit );
      break;
    default:
      m_search_criteria |= Bit;
      break;
    }
  }

  inline Query::bit_flag_type Query::
  Criteria( ) const
  {
    return m_search_criteria;
  }

  inline Query::gps_time_type Query::
  EndTime( ) const
  {
    return m_time_end;
  }

  inline void Query::
  EndTime( gps_time_type Value )
  {
    set_criteria_bit( CRITERIA_BIT_END_TIME );
    m_time_end = Value;
    normalize( );
  }

  inline const std::string&  Query::
  Extension( ) const
  {
    return m_ext;
  }

  inline void Query::
  Extension( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_EXT );
    m_ext = Value;
    normalize( );
  }

  inline Query::filename_handler_type Query::
  FNHandler( ) const
  {
    return m_filename_handler;
  }

  inline void Query::
  FNHandler( filename_handler_type Value )
  {
    set_criteria_bit( CRITERIA_BIT_FILENAME_HANDLER );
    m_filename_handler = Value;
    normalize( );
  }

  inline const std::string& Query::
  IFO( ) const
  {
    return m_ifo;
  }

  inline void Query::
  IFO( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_IFO );
    m_ifo = Value;
    normalize( );
  }

  inline const std::string& Query::
  IFOType( ) const
  {
    return m_ifo_type;
  }

  inline void Query::
  IFOType( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_IFO_TYPE );
    m_ifo_type = Value;
    normalize( );
  }

  inline const std::string& Query::
  IFOTypeAbsolute( ) const
  {
    return m_ifo_type;
  }

  inline void Query::
  IFOTypeAbsolute( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_IFO_TYPE_ABSOLUTE );
    m_ifo_type = Value;
    normalize( );
  }

  inline const Query::ifo_type_list_type& Query::
  IFOTypeList( ) const
  {
    return m_ifo_type_list;
  }

  inline void Query::
  IFOTypeList( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_IFO_TYPE_LIST );

    size_t pos = 0;
    m_ifo_type_list.erase( m_ifo_type_list.begin( ),
			   m_ifo_type_list.end( ) );
    do
    {
      size_t end = Value.find_first_of( " ", pos );

      m_ifo_type_list.push_back( Value.substr( pos, end ) );
      pos = end;
    } while( pos != std::string::npos );
    normalize( );
  }

  inline Query::gps_time_type Query::
  StartTime( ) const
  {
    return m_time_start;
  }

  inline void Query::
  StartTime( gps_time_type Value )
  {
    set_criteria_bit( CRITERIA_BIT_START_TIME );
    m_time_start = Value;
    normalize( );
  }

  inline const std::string& Query::
  MountPoint( ) const
  {
    return m_mount_point;
  }

  inline void Query::
  MountPoint( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_MOUNT_POINT );
    m_mount_point = Value;
    normalize( );
  }

  inline Query::time_range_type& Query::
  TimeRange( )
  {
    if ( m_time_range )
    {
      return *m_time_range;
    }
    throw std::range_error( "Query::TimeRange" );
  }

  inline const Query::time_range_type& Query::
  TimeRange( ) const
  {
    if ( m_time_range )
    {
      return *m_time_range;
    }
    throw std::range_error( "Query::TimeRange" );
  }

  inline void Query::
  TimeRange( gps_time_type Start, gps_time_type Stop )
  {
    set_criteria_bit( CRITERIA_BIT_TIME_RANGE );
    m_time_range = new time_range_type;
    (*m_time_range)[ Start ] = Stop;
    m_time_range_parent = this;
    normalize( );
  }

  inline void Query::
  TimeRange( time_range_type& NewRange )
  {
    if ( m_time_range )
    {
      m_time_range->swap( NewRange );
      // *m_time_range = NewRange;
    }
  }

  inline const std::string& Query::
  Type( ) const
  {
    return m_type;
  }

  inline void Query::
  Type( const std::string& Value )
  {
    set_criteria_bit( CRITERIA_BIT_TYPE );
    m_type = Value;
    normalize( );
  }
} // namespace - diskCache

#endif /* DISKCACHE__QUERY_HH */

