#ifndef FrameCPP_VERSION_3_FrVect_HH
#define FrameCPP_VERSION_3_FrVect_HH

#include <memory>
#include <vector>
#include <string>

#include "general/types.hh"
#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/Compression.hh"
#include "framecpp/Common/FrVect.hh"

#include "framecpp/Version3/FrameSpec.hh"
#include "framecpp/Version3/Dimension.hh"
#include "framecpp/Version3/STRING.hh"

namespace FrameCPP
{
  namespace Version_3
  {
    class FrVectDataTypes
    {
    public:
      enum data_types_type {
	FR_VECT_C = 0,
	FR_VECT_2S = 1,
	FR_VECT_8R = 2,
	FR_VECT_4R = 3,
	FR_VECT_4S = 4,
	FR_VECT_8S = 5,
	FR_VECT_8C = 6,
	FR_VECT_16C = 7,
	FR_VECT_STRING = 8,
	FR_VECT_2U = 9,
	FR_VECT_4U = 10,
	FR_VECT_8U = 11,
	FR_VECT_1U = 12
      };
    };

    class FrVectCompressionScheme
    {
    public:
      static const INT_2U BYTE_SWAP = 0x100;

      enum compression_scheme_type {
	RAW = 0,
	GZIP = 1,
	DIFF = 2,
	DIFF_GZIP = 3
      };
    };

    class FrVect
      : public Common::FrameSpec::Object,
        public Common::FrVect,
        public FrVectDataTypes,
	public FrVectCompressionScheme
    {
    public:
      // using FrameCPP::Compression;

      using Common::FrameSpec::ObjectInterface::Create;

      typedef Common::FrVect::cmn_compression_scheme_type
      cmn_compression_scheme_type;
      typedef Common::FrVect::cmn_compression_level_type
      cmn_compression_level_type;

      typedef INT_2U compress_type;
      typedef INT_2U type_type;
      typedef INT_4U nData_type;
      typedef INT_4U nBytes_type;
      typedef INT_4U nDim_type;

      static const int DEFAULT_GZIP_LEVEL;

      static std::auto_ptr< Object >
      Promote( INT_2U Source,
	       Object* Obj, Common::IStream* Stream );

      static const char* const StructName( );

      static const Common::Description* StructDescription( );

      //: Default constructor
      FrVect( );

      //-----------------------------------------------------------------
      //
      //: Copy Constructor.  
      //
      //!param: const Vect& vect - Copied vector.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const FrVect& vect );


      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+       containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data. Default
      //+       is BIGENDIAN.   
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //    
      FrVect( const std::string& name, type_type type, nDim_type nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      const void* data = 0, 
	      const std::string& unitY = "" );

      //-----------------------------------------------------------------
      //
      //: Constructor.
      //
      //!param: const std::string& name - The name of the data.
      //!param: INT_2U type - The data type.
      //!param: INT_4U nDim - The number of dimensions.
      //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
      //+     containing information about the dimensions for this data.
      //!param: const byte_order_type byte_order - Byte order of the data.
      //+	Default is BIGENDIAN.      
      //!param: const void* data - A pointer to the data.  Default: 0
      //!param: const std::string& unitY - Units for the data.  Default: ""
      //!param: bool allocate - Whether to allocate data. Default: true.
      //!param: bool owns - Whether to destruct data in the destructor.
      //+	Default: true.
      //
      //!exc: std::bad_alloc - Memory allocation failed.
      //
      FrVect( const std::string& name, type_type type, nDim_type nDim,
	      const Dimension* dims,
	      const byte_order_type byte_order = BYTE_ORDER_HOST,
	      void* data = 0,
	      const std::string& unitY = "", 
	      bool allocate = true, bool owns = true );

      FrVect( const std::string& name,
	      compress_type Compress,
	      type_type type,
	      nDim_type nDim,
	      const Dimension* dims,
	      nData_type NData,
	      nBytes_type NBytes,
	      void* data,
	      const std::string& unitY = "", 
	      bool allocate = true,
	      bool owns = true );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, nDim_type nDim = 1,
	      const Dimension* dims = 0,
	      const T* data = 0, const std::string& unitY = "" );

      //!exc: std::bad_alloc - Memory allocation failed.   
      template< class T >
      FrVect( const std::string& name, nDim_type nDim = 1,
	      const Dimension* dims = 0,
	      T* data = 0, const std::string& unitY = "",
	      bool allocate = true, bool owns = true );

      virtual ~FrVect( );

      virtual const char* const ObjectStructName( ) const;

      //: Comparison operator
      bool operator==( const FrVect& RHS ) const;
      bool operator!=( const FrVect& RHS ) const;
      //: Concatination operator
      FrVect& operator+=( const FrVect& RHS );

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      //: Number of bytes needed to write this structure
      FrameCPP::cmn_streamsize_type Bytes( const Common::StreamBase& Stream ) const;

      // Accessors
      const std::string& GetName( ) const;
      compress_type GetCompress( ) const;
      type_type GetType( ) const;
      nData_type GetNData( ) const;
      nBytes_type GetNBytes( ) const;
      const CHAR_U* GetData( bool Decompress = true ) const;
      CHAR_U* GetData( bool Decompress = true );

      const CHAR_U* GetDataRaw( ) const;

      CHAR_U* GetDataRaw( );

      template< class T > static INT_2U GetDataType();
      nDim_type GetNDim( ) const;
      const Dimension& GetDim( nDim_type Offset ) const;
      Dimension& GetDim( nDim_type Offset );
      const std::string GetUnitY( ) const;
      CHAR_U* SteelData( bool& Owns ) const;

      // Mutators
      void Compress( compression_scheme_type Scheme, int GZipLevel );
      void Uncompress( );

      virtual void
      CompressData( cmn_compression_scheme_type Scheme,
		    cmn_compression_level_type GZipLevel );

      virtual Compression::compression_base_type
      Compression( ) const;

      virtual Common::FrameSpec::Object*
      CloneCompressed( cmn_compression_scheme_type Scheme,
		       cmn_compression_level_type Level ) const;

      //-----------------------------------------------------------------
      /// \brief Establish the channel name.
      ///
      /// \param[in] Name
      ///     The channel name
      //-----------------------------------------------------------------
      void SetName( const std::string& Name );

      void SetNData( INT_4U NData );

      size_t GetTypeSize( ) const;
      static size_t GetTypeSize( type_type type );

      //: Merge with another FrAdcData
      FrVect& Merge( const FrVect& RHS );

      std::auto_ptr< FrVect > SubFrVect( INT_4U Start, INT_4U Stop ) const;

      virtual FrVect* Create( Common::IStream& Stream ) const;

      //-----------------------------------------------------------------
      /// \brief Verify the integrity of the object
      ///
      /// \param Verifier
      ///     The handle to the Verify object used to store data that
      ///     needs to be accumulated before being processed.
      /// \param Stream
      ///     The input stream from which the object was created.
      ///
      /// \remarks
      ///     Upon failure, this method thows an exception.
      //-----------------------------------------------------------------
      virtual void VerifyObject( Common::Verify& Verifier,
				 Common::IFrameStream& Stream ) const;

      //: Write the structure to the stream
      virtual void Write( Common::OStream& Stream ) const;

      void
      expandToBuffer( General::AutoArray< CHAR_U >& Dest,
		      INT_8U& DestNBytes ) const;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Common::FrameSpec::Object >
      demote( INT_2U Target, Common::IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Common::FrameSpec::Object >
      promote( INT_2U Target, Object* Obj, Common::IStream* Stream ) const;

    private:
      struct data_type
      {
      public:
	STRING				name;
	compress_type			compress;
	type_type			type;
	nData_type			nData;
	nBytes_type			nBytes;
	CHAR_U*				data;
	std::vector< Dimension >	dimension;
	STRING				unitY;

        data_type();
        data_type( const std::string& n, type_type t, 
		   const byte_order_type byte_order, INT_4U ndim,
		   const Dimension* dims, const std::string& unit );

        data_type( const std::string& Name,
		   compress_type Compress,
		   type_type Type, 
		   nData_type NData,
		   nBytes_type NBytes,
		   void* Data,
		   nDim_type NDim,
		   const Dimension* Dims,
		   const std::string& UnitY,
		   bool Allocate );

	void copy_core( const data_type& Source );

	bool operator==( const data_type& RHS ) const;

      };
 
      static const Common::FrameSpec::Info::frame_object_types s_object_id
      = Common::FrameSpec::Info::FSI_FR_VECT;

      data_type	m_data;
      bool	m_owns;

      FrVect( Common::IStream& Stream );

      void copy_data( CHAR_U* Data, INT_8U size,
		      bool allocate, bool own );

    };


    inline Compression::compression_base_type FrVect::
    Compression( ) const
    {
      return ( 0x0FF & m_data.compress );
    }

    inline std::auto_ptr< Common::FrameSpec::Object > FrVect::
    Promote( INT_2U Source,
	     Common::FrameSpec::Object* Obj, Common::IStream* Stream )
    {
      return Common::PromoteObject( DATA_FORMAT_VERSION, Source, Obj );
    }

    inline const char* const FrVect::
    StructName( )
    {
      static const CHAR* class_name( "FrVect" );
      return class_name;
    }

    inline const CHAR_U* FrVect::
    GetDataRaw( ) const
    {
      return m_data.data;
    }

    inline CHAR_U* FrVect::
    GetDataRaw( )
    {
      return m_data.data;
    }

    inline FrVect::compress_type FrVect::
    GetCompress( ) const
    {
      return m_data.compress;
    }

    inline FrVect::type_type FrVect::
    GetType( ) const
    {
      return m_data.type;
    }

    inline size_t FrVect::
    GetTypeSize( ) const
    {
      return GetTypeSize( m_data.type );
    }

    inline FrVect::nBytes_type FrVect::
    GetNBytes( ) const
    {
      return m_data.nBytes;
    }

    inline FrVect::nDim_type FrVect::GetNDim( ) const
    {
      return m_data.dimension.size( );
    }

    inline const Dimension& FrVect::
    GetDim( nDim_type Offset ) const
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline Dimension& FrVect::
    GetDim( nDim_type Offset )
    {
      //:TODO: Throw exception if value out of range
      return m_data.dimension[ Offset ];
    }

    inline const std::string FrVect::
    GetUnitY( ) const
    {
      return m_data.unitY;
    }

    //-------------------------------------------------------------------
    /// Assign a new value for the channel name.
    //-------------------------------------------------------------------
    inline void FrVect::
    SetName( const std::string& Name )
    {
      m_data.name = Name;
    }

    inline void FrVect::
    SetNData( nData_type NData )
    {
      m_data.nData = NData;
    }
   
    inline CHAR_U* FrVect::
    SteelData( bool& Owns ) const
    {
      Owns = m_owns;
      const_cast< FrVect* >( this )->m_owns = false;
      return m_data.data;
    }
  } // namespace - Version_3
} // namespace FrameCPP

#if defined( __SUNPRO_CC ) && ( __SUNPRO_CC <= 0x550 )
#include "framecpp/Version3/FrVect.tcc"
#endif /* */

#endif /* FrameCPP_VERSION_3_FrVect_HH */
